# PRG - Exam #

This is the assignment for the PRG exam. Follow the instructions precisely and please note:

- If the module executes any functions when imported (e.g. print), points will be subtracted.
- If the module fails to be imported (due to syntactic errors), the solution is invalid.

The modules (packages) you implement should be packed in a zip file (files should be in the root of
the zip file, zipping directly the files, not the containing folder) and uploaded to BRUTE. For
every module, class and function you implement, do not forget to add a docstring. Code quality
(mainly following naming conventions and writing docstrings) is part of the grading of your
solution.


## Task 1 [3p] ##

Create a module named "variables" which will define 6 pairs of variables of builtin types with
the following names and properties:

- different_type_1, different_type_2: each variable has a different type
- children_builtin_type, parent_builtin_type: children_type variable is a subclass of
        parent_type and both are builtin types. In other words, the children_type variable is also
        an instance of the type of variable parent_type
- different_value_1, different_value_2: the value of variables is different, but their data type
        is the same
- same_heap_1, same_heap_2: the data on the heap is shared between the variables (i.e. they have
        the same memory address), but their data type is immutable
- alias_1, alias_2: the data on the heap is shared and the two variables are mutable
- different_heap_1, different_heap_2: the two variables are equal, but they do not share data
        on the heap (each has a different address).


## Task 2 [5p] ##

Create a module named "task2_functions" and implement three functions according to the following
unit tests. You can copy the unit tests to the same module.

```
import unittest

class TestFunctions(unittest.TestCase):

    def test_is_palindrome(self):
        # [1p]
        self.assertTrue(is_palindrome("121"))
        self.assertTrue(is_palindrome("Anna"))
        self.assertTrue(is_palindrome("Step on no pets"))
        self.assertFalse(is_palindrome("123"))
        self.assertFalse(is_palindrome("Anne"))
        self.assertFalse(is_palindrome("Step on pets"))

    def test_format_balance(self):
        # [1.5p]
        self.assertEqual("Your balance is 0 EUR", format_balance("EUR", []))
        self.assertEqual("Your balance is 20 USD", format_balance("USD", [15, 5]))
        self.assertEqual("Your balance is 100 GBP", format_balance("GBP", [125, -15, -10]))
        self.assertEqual("You have a debt of 200 CZK", format_balance("CZK", [-200]))

    def test_beep_banned_words(self):
        # [2.5p]
        self.assertEqual("My **** is Johana", beep_banned_words("My name is Johana", ["name"]))
        self.assertEqual("This is ******* awesome", beep_banned_words("This is greatly awesome",
                         ["greatly"]))
        self.assertEqual("**** ****", beep_banned_words("Long talk", ["long", "talk"]))
```


## Task 3 [1p] ##

Create a package with a single module containing the class definition below.

1. [1p] Choose the name of the package and module appropriately. Be sure to follow best practices.

```
class PRGPoints:
    """A sequence (list) of points for a single student for the course of Programming Essentials."""

    course = "Programming Essentials"

    def __init__(self, student_name, points):
        self.student_name = student_name
        self.points = points

    def __repr__(self):
        points_str = []
        for point in self.points:
            points_str.append(str(point))
        return f"{self.student_name}: {', '.join(points_str)}"
```


## Task 3.1 [6p] ##

In the module created in Task 3, implement the following tasks.

1. [1p] Add a method named "passed_semester" that will return a boolean whether the total number
    of points (their sum) is greater or equal to 30.
2. [1p] Add support for the equality operator "==". Instances are equal if all their instance
    variables are equal.
3. [2p] Add support for adding two instances using the addition operator "+". Instances can be added
    only if the student name is the same. The resulting instance should contain a concatenated
    sequence of points.
4. [2p] Write a unit test for the addition operator and put it in a test case named "TestPRGPoints".


## Task 3.2 [5p] ##

In the module created in Task 3, implement the following tasks.

1. [2p] Write function "save_points" which takes two arguments - a PRGPoints instance and a file
    path, and will store the data from the instance. The structure of the file is up to you.
2. [3p] Write function "load_points" which takes a single argument - a file path, and returns
    a PRGPoints instance containing the data previously saved by save_points.


## Task 3.3 [5p] ##

In the module created in Task 3, implement the following tasks.

1. [1p] Implement class ALGPoints which will inherit from PRGPoints but have the course class
    variable equal to "Algorithms".
2. [2p] Make sure that each element in the points argument of initializer is bigger than 5.
    Otherwise, raise a custom exception that you will implement.
3. [2p] Extend functionality of "save_points" and "load_points" to support ALGPoints instances
    as well. The "load_points" function should return the same class that was saved
    by "save_points". The structure of the file is up to you; do not change the arguments
    of the functions.


## Epilogue ##

At the end of the exam, the following files and folders should be present in your zip file. They
should be in the root of the zip file, not in a nested folder.

- variables.py - task 1
- task2_functions.py - task 2
- [folder of chosen name] - task 3
